#' self-X autonomic unsupervised feature selection
#' @description
#' With the aim of making UNSUPERVISED feature selection more autonomic,
#' an ad-hoc procedure is implemented here following next approach:
#' \enumerate{
#' \item Initially, well-known techniques as PCA and FA are used
#' to select the most important features under different criteria.
#' \item Subsequently, an OWA operator is applied to to determine which
#'  3 variables are the most relevant to be used as input in the following
#'  clustering phase
#' }
#' @keywords package selfX s-X-AIPI autonomic feature selection
#' @details
#' The feature selection process applied to the data is as follows:
#' \itemize{
#' \item Missing values are treated (assigned to zero).
#' \item The data is normalized (Z-Score standarization).
#' \item Different unsupervised feature selection techniques are applied to the data
#'   to establish different importance values of the variables considered.
#'
#'   Specifically 4 methods:
#'
#'    \itemize{
#'     \item Evaluation with PCA. 3 variants of the method are applied:
#'      \itemize{
#'        \item (1) PCA on the first principal component.
#'        \item (2) PCA on all those principal components that are above the mean eigenvalues.
#'        \item (3) PCA on all principal components that contain 90\% of the explained variance.
#'      }
#'     In all cases, the importance value of the variable is assigned based on the
#'     sum of the loadings of the selected principal components, normalized by the
#'     total sum of all loadings.
#'
#'     \item Evaluation with FA. 1 variant of the method is applied:
#'         \itemize{
#'         \item (4) FA on all factors. To determine the number of factors, Horn's
#'         Parallel Analysis is applied. A Maximum Neighborhood factorization
#'         and Varimax rotation are used.
#'         }
#'     In all cases, the importance value of the variable is assigned based on
#'     the \emph{communality} value of the variable, normalized by the sum of all
#'     \emph{communalities}.
#'     }
#' \item Based on the evaluation obtained, a multi-criteria decision-making
#'   process is carried out based on:
#'     \itemize{
#'        OWA (Ordered Weighted Averaging) operator: applying the RIM (Regular
#'        Increasing Monotone) quantifier \eqn{\alpha(r)=r^{\alpha}},
#'        \eqn{r\geq 0} with \eqn{orness=0.4} on the evaluations, an importance is
#'        established for each characteristic, selecting in each case only
#'        the most important ones.
#'     }
#' \item A subset of features is selected and returned.
#' }
#' @param dataDF dataframe containing clean data (no outliers,
#' no missing values, no constant values) for data mining activities.
#' ONLY numerical input variables allowed (that it can be normalized or not).
#' @return text string with the names of the selected variables
#' @references
#'
#' A. Malhi. "PCA-based feautre selection scheme for machine defect classification"
#' IEEE Transactions on Instrumentation and Measurement, 35(6):1517–1525, 2004.
#'
#' K. Z. Mao. "Identifying critical variables of principal components for
#' unsupervised feature selection" IEEE Transaction on System, Man
#' and Cybernetics, 35(2):339–344, 2005.
#'
#' W.J. Krzanowski and F.H.C Marriott. Multivariate analysis (2 vols).
#' Edward Arnold, 1994.
#'
#' R.R. Yager. "On ordered weighted averaging aggregation operators in
#' multi-criteria decision making" IEEE Transactions on Systems, Man
#' and Cybernetics, 18:183–190, 1998.
#'
#' M. Galende-Hernández, M. Menéndez, M.J. Fuente, G.I. Sainz-Palmero
#' "Monitor-While-Drilling-based estimation of rock mass rating with
#' computational intelligence: The case of tunnel excavation front"
#' Automation in Construction, 93:325-338, 2018.
#'
#' @note
#' The s-X-AIPI project has received funding from the European Union's Horizon
#' Europe research and innovation programme under grant agreement Nº 101058715
#' \url{https://s-x-aipi-project.eu/}
#' @seealso \code{\link[stats:prcomp]{prcomp}} for PCA, \code{\link[psych:fa.parallel]{fa.parallel}} for FA and \code{\link[agop:owa]{owa}} for OWA.
#' @examples
#' ## load dataset
#' data(mtcars)
#' ## run pca.fa.owa method
#' pca.fa.owa(mtcars)
#' @export
pca.fa.owa <- function(dataDF) {

  DF_SelecCaracteristicas <- dataDF

  optionSC = 1
  # optionSC = 1 --> se ejecuta PCA + FA
  # optionSC = 2 --> se ejecuta PCA

  debugger <- TRUE
  debugger <- FALSE

  watchful <- list(cod = 0, txt = "")

  # COMPROBAR LOS PARAMETROS DE ENTRADA
  if (is.null(DF_SelecCaracteristicas) | nrow(DF_SelecCaracteristicas) == 0 | ncol(DF_SelecCaracteristicas) == 0) {
    watchful$cod <- 1
    watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: data not available", sep = " ")
    watchful$txt <- paste(watchful$txt, "", sep = "\n")
  }
  # necesita vbles numericas sin NAs
  DF_SelecCaracteristicas[is.na(DF_SelecCaracteristicas)] <- 0 # Cambiar NAs por ceros antes de aplicar PCA
  summary(DF_SelecCaracteristicas)

  todas_numericas <- all(sapply(DF_SelecCaracteristicas, is.numeric))
  if (!todas_numericas) {
    watchful$cod <- 1
    watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: not numerical data", sep = " ")
    watchful$txt <- paste(watchful$txt, "", sep = "\n")
  }
  rm(todas_numericas)


  if (watchful$cod == 0) {

    DF_SelecCaracteristicas <- data.frame(DF_SelecCaracteristicas)

    # 1 - Normalizar datos ZSCORE vbles numericas
    DF_SelecCaracteristicas.norm <- clusterSim::data.Normalization(DF_SelecCaracteristicas[], type = "n1", normalization = "column")
    summary(DF_SelecCaracteristicas.norm)

    nvbles <- length(DF_SelecCaracteristicas.norm)

    ############################ PCA ############################
    # Variables generales de la SC
    if (optionSC == 1) {
      nIndices <- 4 # PCA1,PCA2,PCA3,FA
    } else {
      if (optionSC == 2) {
        # SIN FA
        nIndices <- 3 # PCA1,PCA2,PCA3 --> modificar tb 1+3 lineas mas adelante (al calcular OWA)
      } else {
        return("ERROR")
      }
    }
    nOrnessOWA <- 3 # ornes=0.3,0.4,0.9
    orness <- c(0.3, 0.4, 0.9)
    IndicesPCAFA <- matrix(nrow = nvbles, ncol = (nIndices * 2) + (nOrnessOWA * 2))
    if (optionSC == 1) {
      colnames(IndicesPCAFA) <- c("Eval1PCA", "rank1PCA", "Eval2PCA", "rank2PCA", "Eval3PCA", "rank3PCA", "EvalFA", "rankFA", "OWAvalue03", "OWApor03", "OWAvalue04", "OWApor04", "OWAvalue09", "OWApor09")
    } else {
      if (optionSC == 2) {
        # SIN FA
        colnames(IndicesPCAFA) <- c("Eval1PCA", "rank1PCA", "Eval2PCA", "rank2PCA", "Eval3PCA", "rank3PCA", "OWAvalue03", "OWApor03", "OWAvalue04", "OWApor04", "OWAvalue09", "OWApor09")
      }
    }
    rownames(IndicesPCAFA) <- colnames(DF_SelecCaracteristicas.norm)

    summary(DF_SelecCaracteristicas.norm)
    # pca <- princomp(DF_SelecCaracteristicas.norm)
    pca <- stats::prcomp(DF_SelecCaracteristicas.norm, scale = TRUE)
    pvar <- (pca$sdev^2) / sum(pca$sdev^2)
    cvar <- cumsum(pvar)

    # PCA sobre la primera componente principal
    # pesoXvble <- abs(pca$loading[,1])
    pesoXvble <- abs(pca$rotation[, 1])
    pesoXvbleN_1PC <- pesoXvble / sum(pesoXvble) # sum(pesoXvbleN_1PC) es 1
    rankXvbleN_1PC <- rank(pesoXvbleN_1PC) # asignar rankings. 0 si peso norma 0; 1 si meso norma  maximo

    IndicesPCAFA[, 1] <- pesoXvbleN_1PC
    IndicesPCAFA[, 2] <- rankXvbleN_1PC

    # PCA sobre todas aquellas componentes principales que estan por encima de la media de los autovalores
    nCP <- max(which(pca$sdev > mean(pca$sdev)))
    if (nCP == 1) pesoXvble <- abs(pca$rotation[, 1])
    if (nCP != 1) pesoXvble <- rowSums(abs(pca$rotation[, 1:nCP]))
    # if (nCP==1) pesoXvble <- abs(pca$loading[,1])
    # if (nCP!=1) pesoXvble <- rowSums(abs(pca$loading[,1:nCP]))
    pesoXvbleN_Media <- pesoXvble / sum(pesoXvble) # sum(pesoXvbleN_90) es 1
    rankXvbleN_Media <- rank(pesoXvbleN_Media) # asignar rankings. 0 si peso norma 0; 1 si meso norma  maximo

    IndicesPCAFA[, 3] <- pesoXvbleN_Media
    IndicesPCAFA[, 4] <- rankXvbleN_Media

    # PCA sobre todas las componentes principales que contienen 90% de la varianza acumulada
    nCP <- min(which(cvar >= 0.9))
    # if (nCP==1) pesoXvble <- abs(pca$loading[,1])
    # if (nCP!=1) pesoXvble <- rowSums(abs(pca$loading[,1:nCP]))
    if (nCP == 1) pesoXvble <- abs(pca$rotation[, 1])
    if (nCP != 1) pesoXvble <- rowSums(abs(pca$rotation[, 1:nCP]))
    pesoXvbleN_90 <- pesoXvble / sum(pesoXvble) # sum(pesoXvbleN_90) es 1
    rankXvbleN_90 <- rank(pesoXvbleN_90) # asignar rankings. 0 si peso norma 0; 1 si meso norma  maximo

    IndicesPCAFA[, 5] <- pesoXvbleN_90
    IndicesPCAFA[, 6] <- rankXvbleN_90

    rm(pca, pvar, cvar)
    rm(pesoXvble, nCP, pesoXvbleN_1PC, rankXvbleN_1PC, pesoXvbleN_Media, rankXvbleN_Media, pesoXvbleN_90, rankXvbleN_90)

    ############################ FA  ############################

    # # Para determinar nFactores se usa Analisis Paralelo de Horn. La grafica que se genera se almacena en fichero
    # FileName <- paste("/FactorAnalysis.png",sep="")
    # png(FileName)
    # Obtener el nombre del sistema operativo
    sistema_operativo <- Sys.info()["sysname"]
    # Comprobar si el sistema operativo es Windows
    if (sistema_operativo == "Windows") {
      sink("nul")
      nFactoresFA <- psych::fa.parallel(DF_SelecCaracteristicas.norm, fm = "ml", fa = "both", plot = FALSE)
      sink()
    } else if (sistema_operativo == "Linux") { # Comprobar si el sistema operativo es Linux
      sink("/dev/null") # now suppresses
      nFactoresFA <- psych::fa.parallel(DF_SelecCaracteristicas.norm, fm = "ml", fa = "both", plot = FALSE)
      sink()
    } else {
      nFactoresFA <- psych::fa.parallel(DF_SelecCaracteristicas.norm, fm = "ml", fa = "both", plot = FALSE)
    }
    # # Parallel analysis suggests that the number of factors =  5  and the number of components =  3
    # dev.off()
    # rm(FileName)

    if (nFactoresFA$nfact > 0) {
      fa <- psych::fa(DF_SelecCaracteristicas.norm, nFactoresFA$nfact, rotate = "varimax", fm = "ml")

      if (nFactoresFA$nfact == 1) {
        pesoXvble <- fa$loading[, 1:nFactoresFA$nfact]^2
      } else {
        pesoXvble <- rowSums(fa$loading[, 1:nFactoresFA$nfact]^2)
      }
      pesoXvbleN_FA <- pesoXvble / sum(pesoXvble) # sum(pesoXvbleN_90) es 1
      rankXvbleN_FA <- rank(pesoXvbleN_FA) # asignar rankings. 0 si peso norma 0; 1 si peso norma  maximo
      rm(pesoXvble, fa)
    } else {
      # print("No data to run FA")
      pesoXvbleN_FA <- rep(1, nvbles)
      rankXvbleN_FA <- rank(pesoXvbleN_FA)
    }

    IndicesPCAFA[, 7] <- pesoXvbleN_FA
    IndicesPCAFA[, 8] <- rankXvbleN_FA

    rm(nFactoresFA, pesoXvbleN_FA, rankXvbleN_FA)

    ############################# OWA sobre las evaluaciones (no sobre los rankings)  ############################

    for (j in 1:nOrnessOWA) {
      operadorOR <- orness[j]
      # Calculo Pesos
      alpha <- (1 - operadorOR) / operadorOR
      w <- matrix()
      for (i in 1:nIndices) {
        w[i] <- ((i / nIndices)^alpha) - ((i - 1) / nIndices)^alpha
      }
      rm(i, alpha)

      # Relleno tabla
      for (i in 1:nvbles) {
        if (optionSC == 1) {
          IndicesPCAFA[i, (nIndices * 2) + (j * 2) - 1] <- agop::owa(IndicesPCAFA[i, c(1, 3, 5, 7)], w)
        } else {
          # Si no realiza el analisis FA hay que quitar el 7 del final
          IndicesPCAFA[i, (nIndices * 2) + (j * 2) - 1] <- agop::owa(IndicesPCAFA[i, c(1, 3, 5)], w)
        }
      }
      rm(i)
    }
    rm(j, operadorOR)

    if (optionSC == 1) {
      # #  calcular % segun owa columna 9 para columna 10
      IndicesPCAFA[, 10] <- (IndicesPCAFA[, 9] / sum(IndicesPCAFA[, 9])) * 100
      IndicesPCAFA[, 12] <- (IndicesPCAFA[, 11] / sum(IndicesPCAFA[, 11])) * 100
      IndicesPCAFA[, 14] <- (IndicesPCAFA[, 13] / sum(IndicesPCAFA[, 13])) * 100
    } else {
      # Si no realiza el analisis FA hay que descomentar estas lineas y comentar las anteriores
      IndicesPCAFA[, 8] <- (IndicesPCAFA[, 7] / sum(IndicesPCAFA[, 7])) * 100
      IndicesPCAFA[, 10] <- (IndicesPCAFA[, 9] / sum(IndicesPCAFA[, 9])) * 100
      IndicesPCAFA[, 12] <- (IndicesPCAFA[, 11] / sum(IndicesPCAFA[, 11])) * 100
    }

    IndicesPCAFA <- as.data.frame(IndicesPCAFA)
    if (debugger) print(IndicesPCAFA)

    rm(nvbles, nIndices, w, nOrnessOWA, orness)
    rm(DF_SelecCaracteristicas.norm)

    # watchful$txt <- varsINPUTselected
    varsINPUTselected <- row.names(IndicesPCAFA[order(-IndicesPCAFA$OWApor04), ])[1:3]

    rm(IndicesPCAFA)

    cat("Self-X UnSupervised Feature Selection: ", varsINPUTselected, "\n", sep = "\t")
    return(varsINPUTselected)
  } else {
    if (watchful$cod == 1) {
      cat(watchful$txt)
      return(0)
    }
  }
} #pca.fa.owa
